from abcli.plugins.tags.functions import get, set_
from abcli.modules.objects import unique_object


def test_abcli_tags():
    object_name = unique_object()

    assert set_(object_name, "this,that")

    tags_as_read = get(object_name)
    assert "this" in tags_as_read
    assert "that" in tags_as_read



import argparse
from abcli.plugins.metadata import get, NAME, post, MetadataSourceType
from abcli.logger import logger
from blueness.argparse.generic import sys_exit


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="get|post",
)
parser.add_argument(
    "--default",
    type=str,
    default="",
)
parser.add_argument(
    "--delim",
    type=str,
    default="",
)
parser.add_argument(
    "--dict_keys",
    type=int,
    default=0,
    help="0|1",
)
parser.add_argument(
    "--dict_values",
    type=int,
    default=0,
    help="0|1",
)
parser.add_argument(
    "--filename",
    type=str,
    default="metadata.yaml",
)
parser.add_argument(
    "--is_base64_encoded",
    type=int,
    default=0,
    help="0|1",
)
parser.add_argument(
    "--key",
    type=str,
)
parser.add_argument(
    "--source",
    type=str,
)
parser.add_argument(
    "--source_type",
    type=str,
    default=MetadataSourceType.FILENAME.name.lower(),
    help="|".join([source_type.name.lower() for source_type in MetadataSourceType]),
)
parser.add_argument(
    "--value",
    type=str,
)
parser.add_argument(
    "--verbose",
    type=int,
    default="0",
    help="0|1",
)
args = parser.parse_args()

delim = " " if args.delim == "space" else args.delim

success = False
if args.task == "get":
    success = True
    output = get(
        key=args.key,
        default=args.default,
        source=args.source,
        source_type=MetadataSourceType[args.source_type.upper()],
        dict_keys=args.dict_keys,
        dict_values=args.dict_values,
        filename=args.filename,
    )
    print(delim.join(output) if isinstance(output, list) else output)
elif args.task == "post":
    success = post(
        key=args.key,
        value=args.value,
        filename=args.filename,
        source=args.source,
        source_type=MetadataSourceType[args.source_type.upper()],
        is_base64_encoded=args.is_base64_encoded,
        verbose=args.verbose == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



import numpy as np


def add_frame(image, width):
    """add frame to image.

    Args:
        image (np.ndarray): image.
        width (int): frame width.

    Returns:
        np.ndarray: updated image.
    """
    output = np.zeros(
        (image.shape[0] + 2 * width, image.shape[1] + 2 * width, image.shape[2]),
        dtype=image.dtype,
    )

    output[width:-width, width:-width, :] = image[:, :, :]

    return output



from abcli import env
import logging
from logging.handlers import RotatingFileHandler

# to hide "botocore.credentials Found credentials in environment variables."
logging.getLogger("botocore.credentials").setLevel(logging.WARNING)

ABCUL = " \\\n\t"

abcli_log_filename = env.abcli_log_filename

# Based on https://stackoverflow.com/a/22313803
logging.addLevelName(logging.INFO, "")
logging.addLevelName(logging.DEBUG, "❓ ")
logging.addLevelName(logging.ERROR, "❗️ ")
logging.addLevelName(logging.WARNING, "⚠️  ")

logging_level = logging.INFO

logging.getLogger().setLevel(logging_level)

log_formatter = logging.Formatter("%(levelname)s%(name)s %(message)s")
try:
    file_handler = RotatingFileHandler(
        abcli_log_filename,
        maxBytes=10485760,
        backupCount=10000,
    )
    file_handler.setLevel(logging_level)
    file_handler.setFormatter(log_formatter)
    logging.getLogger().addHandler(file_handler)
except:
    pass

console_handler = logging.StreamHandler()
console_handler.setLevel(logging_level)
console_handler.setFormatter(log_formatter)
logging.getLogger().addHandler(console_handler)

logger = logging.getLogger("::")


def crash_report(description):
    # https://stackoverflow.com/a/10645855
    logger.error(f"crash: {description}", exc_info=1)


def get_logger(ICON):
    return logging.getLogger(f"{ICON} ")



from typing import List
import os
from blueness import module
from abcli import NAME as MY_NAME, ICON as MY_ICON
from abcli import fullname
from abcli import file
from abcli.plugins import markdown
from abcli.logger import logger

MY_NAME = module.name(__file__, MY_NAME)


def build(
    NAME: str,
    VERSION: str,
    REPO_NAME: str,
    items: List[str] = [],
    template_filename: str = "",
    filename: str = "",
    path: str = "",
    cols: int = 3,
    ICON: str = "",
) -> bool:
    if path:
        template_filename = os.path.join(path, "template.md")
        filename = os.path.join(path, "README.md")

    logger.info(
        "{}.build: {} -{}-{}-@-{}-#{}-> {}".format(
            MY_NAME,
            template_filename,
            NAME,
            VERSION,
            REPO_NAME,
            cols,
            filename,
        )
    )

    table = markdown.generate_table(items, cols=cols)

    signature = [
        f'to use on [AWS SageMaker](https://aws.amazon.com/sagemaker/) replace `<plugin-name>` with "{NAME}" and follow [these instructions](https://github.com/kamangir/notebooks-and-scripts/blob/main/SageMaker.md).',
        "",
        " ".join(
            [
                f"[![pylint](https://github.com/kamangir/{REPO_NAME}/actions/workflows/pylint.yml/badge.svg)](https://github.com/kamangir/{REPO_NAME}/actions/workflows/pylint.yml)",
                f"[![pytest](https://github.com/kamangir/{REPO_NAME}/actions/workflows/pytest.yml/badge.svg)](https://github.com/kamangir/{REPO_NAME}/actions/workflows/pytest.yml)",
                f"[![bashtest](https://github.com/kamangir/{REPO_NAME}/actions/workflows/bashtest.yml/badge.svg)](https://github.com/kamangir/{REPO_NAME}/actions/workflows/bashtest.yml)",
                f"[![PyPI version](https://img.shields.io/pypi/v/{REPO_NAME}.svg)](https://pypi.org/project/{REPO_NAME}/)",
            ]
        ),
        "",
        "built by {} [`{}`]({}), based on {}[`{}-{}`]({}).".format(
            MY_ICON,
            fullname(),
            "https://github.com/kamangir/awesome-bash-cli",
            f"{ICON} " if ICON else "",
            NAME,
            VERSION,
            f"https://github.com/kamangir/{REPO_NAME}",
        ),
    ]

    success, template = file.load_text(template_filename)
    if not success:
        return success

    content: List[str] = []
    for template_line in template:
        content_section: List[str] = [template_line]

        if "--table--" in template_line:
            content_section = table
        elif "--signature" in template_line:
            content_section = signature
        elif "--include--" in template_line:
            include_filename_relative = template_line.split(" ")[1].strip()
            include_filename = file.absolute(
                include_filename_relative,
                file.path(template_filename),
            )

            success, content_section = file.load_text(include_filename)
            if not success:
                return success

            content_section = [
                line for line in content_section if not line.startswith("used by:")
            ]

            include_title = (template_line.split(" ", 2) + ["", "", ""])[2]
            if include_title:
                content_section = [f"## {include_title}"] + content_section[1:]

            content_section += [
                "using [{}]({}).".format(
                    file.name(include_filename),
                    include_filename_relative,
                )
            ]

            logger.info(f"{MY_NAME}.build: including {include_filename} ...")

        content += content_section

    return file.save_text(filename, content)



import datetime
from datetime import timezone
import json
import math
import random
import string
import time
from blueness import module
from abcli import NAME
from abcli.string.constants import unit_of

NAME = module.name(__file__, NAME)


def after(s, sub_string, n=1):
    """return what is after sub_string in s.

    Args:
        s (str): string.
        sub_string (str): sub string.
        n (int, optional): instance index. Defaults to 1.

    Returns:
        str: string after the n-th instance of sub_string in s.
    """
    if sub_string == "" or sub_string not in s:
        return ""

    return sub_string.join(s.split(sub_string)[n:])


def before(s, sub_string, n=1):
    """return what is before sub_string in s.

    Args:
        s (str): string.
        sub_string (str): sub string.
        n (int, optional): instance index. Defaults to 1.

    Returns:
        str: string before the n-th instance of sub_string in s.
    """
    if sub_string == "" or sub_string not in s:
        return ""

    return sub_string.join(s.split(sub_string)[:n])


def between(s, sub_string_1, sub_string_2):
    """return what is between sub_string_1 and sub_string2 in string.

    Args:
        s (str): string.
        sub_string_1 (str): sub string.
        sub_string_2 (str): sub string.

    Returns:
        str: what is between sub_string_1 and sub_string2 in string.
    """
    return before(after(s, sub_string_1), sub_string_2)


def pretty_bytes(byte_count):
    """describe byte_count as a string.

    Args:
        byte_count (int): number of bytes.

    Returns:
        str: description of byte_count.
    """
    if byte_count is None:
        return "Unknown"

    byte_count = round(byte_count)

    # Less than a kB
    if byte_count < 1024 / 10:
        return f"{byte_count:.0f} byte(s)"

    byte_count /= 1024

    if byte_count < 1024 / 10:
        return f"{byte_count:.2f} kB"

    byte_count /= 1024

    if byte_count < 1024:
        return f"{byte_count:.2f} MB"

    byte_count /= 1024

    if byte_count < 1024:
        return f"{byte_count:.2f} GB"

    byte_count /= 1024

    return f"{byte_count:.2f} TB"


def pretty_date(
    date=None,
    as_filename=False,
    delimiter=", ",
    explicit_format="",
    in_gmt=False,
    include_date=True,
    include_seconds=True,
    include_time=True,
    include_weekdays=False,
    include_zone=False,
    short=False,
    squeeze=False,
    unique=False,
):
    """return date as a string.

    Args:
        date (Any, optional): date. Defaults to None.
        as_filename (bool, optional): output is to be used as a filename.. Defaults to False.
        delimiter (str, optional): delimiter between date and time. Defaults to ", ".
        explicit_format (str, optional): explicit format, supersedes other settings. Defaults to "".
        in_gmt (bool, optional): in gmt timezone. Defaults to False.
        include_date (bool, optional): include date. Defaults to True.
        include_seconds (bool, optional): show seconds. Defaults to True.
        include_time (bool, optional): include time. Defaults to True.
        include_weekdays (bool, optional): include weekday name. Defaults to False.
        include_zone (bool, optional): show time zone. Defaults to False.
        short (bool, optional): short format. Defaults to False.
        squeeze (bool, optional): squeeze output. Defaults to False.
        unique (bool, optional): add unique postfix. Defaults to False.

    Returns:
        str: date as a string.
    """
    format = ""
    if include_date:
        if short or as_filename:
            format = "%Y %m %d"
        else:
            format = "%d %B %Y"

        if include_weekdays:
            format = "%A, " + format

        if include_time:
            format += delimiter
    if include_time:
        format += "%H:%M"
        if include_seconds:
            format += ":%S"
    if unique:
        format += f":{random.randrange(100000):05d}"
    if as_filename:
        format = (
            format.replace(" ", "-")
            .replace(":", "-")
            .replace(".", "-")
            .replace(",", "")
        )

    if explicit_format:
        format = explicit_format

    if date is None:
        date = time.time()
    if isinstance(date, datetime.datetime):
        output = date.strftime(format)
    else:
        if in_gmt:
            output = time.strftime(format, time.gmtime(date))

            if include_zone:
                output += " (GMT)"
        else:
            output = time.strftime(format, time.localtime(date))

            if include_zone:
                output += f" ({time.tzname[0] if time.tzname else '?'})"

    if squeeze:
        output = output.replace("-", "")

    return output


def pretty_frequency(frequency):
    """describe frequency as a string.

    Args:
        frequency (float): frequency

    Returns:
        str: description of frequency.
    """
    if frequency is None:
        return "Unknown"

    if frequency >= 0.5:
        if frequency < 10**3:
            return f"{frequency:.1f} Hz"

        if frequency < 10**6:
            return f"{frequency / 10**3:.1f} kHz"

        if frequency < 10**6:
            return f"{frequency / 10**6:.1f} MHz"

        return f"{frequency / 10**9:.1f} GHz"

    return "1/{}".format(
        pretty_duration(
            1 / frequency,
            largest=True,
            short=True,
        )
    )


def pretty_param(param, value=None) -> str:
    if isinstance(param, str):
        return unit_of.get(param.split(".")[0], "{}").format(value)

    if isinstance(param, dict):
        return [
            f"{param_}: {pretty_param(param_, value_)}"
            for param_, value_ in param.items()
        ]

    print(f"{NAME}.pretty_param({param.__class__.__name__}), class not found.")
    return ""


def pretty_shape(shape):
    """describe shape as a string.

    Args:
        shape (List[int]): shape.

    Returns:
        str: shape as string.
    """
    return "x".join([str(value) for value in list(shape)])


def pretty_shape_of_matrix(matrix):
    """describe size of matrix as a string.

    Args:
        matrix (Any): matrix.

    Returns:
        str: size of matrix.
    """
    import numpy as np

    return (
        f"{pretty_shape(matrix.shape)}:{matrix.dtype}"
        if isinstance(matrix, np.ndarray)
        else "-"
    )


def pretty_duration(
    duration,
    element_format="{}{}",
    include_ms=False,
    largest=False,
    past=False,
    short=False,
):
    """describe duration as a string.

    Args:
        duration (int): duration in seconds.
        element_format (str, optional): element format. Defaults to "{}{}".
        largest (bool, optional): show the largest element. Defaults to False.
        include_ms (bool, optional): include milliseconds. Defaults to False.
        past (bool, optional): past time, add "ago" to the string. Defaults to False.
        short (bool, optional): produce short description. Defaults to False.

    Returns:
        str: description of duration as a string.
    """
    if duration is None:
        return "None"

    negative_duration = duration < 0
    duration = abs(duration)

    duration_ = duration
    duration = math.floor(duration) if include_ms else round(duration)

    milliseconds = round(1000 * (duration_ - duration))

    seconds = duration % 60
    duration = math.floor(duration / 60)
    minutes = duration % 60
    duration = math.floor(duration / 60)
    hours = duration % 24
    duration = math.floor(duration / 24)
    days = duration % 30
    duration = math.floor(duration / 30)
    months = duration % 12
    years = math.floor(duration / 12)

    if short:
        year_name = " yr"
        month_name = " mth"
        day_name = " d"
        hour_name = " hr"
        minute_name = " min"
        second_name = " s"
        millisecond_name = " ms"
    else:
        year_name = " year(s)"
        month_name = " month(s)"
        day_name = " day(s)"
        hour_name = " hour(s)"
        minute_name = " minute(s)"
        second_name = " second(s)"
        millisecond_name = " millisecond(s)"

    output = []
    if years > 0:
        output.append(element_format.format(years, year_name))
    if months > 0:
        output.append(element_format.format(months, month_name))
    if days > 0:
        output.append(element_format.format(days, day_name))
    if hours > 0:
        output.append(element_format.format(hours, hour_name))
    if minutes > 0:
        output.append(element_format.format(minutes, minute_name))
    if seconds > 0:
        output.append(element_format.format(seconds, second_name))
    if include_ms:
        if milliseconds > 0:
            output.append("{:03d}{}".format(milliseconds, millisecond_name))

    if largest:
        output = output[:1]
    output = ", ".join(output)

    if past and output:
        output += " ago"

    if not output:
        output = "< 1{}".format(millisecond_name if include_ms else second_name)

    return ("-" if negative_duration else "") + output


def random_(
    length=16,
    alphabet=string.ascii_lowercase + string.digits + string.ascii_uppercase,
):
    """generate random string.

    Args:
        length (int, optional): length. Defaults to 16.
        alphabet (str, optional): alphabet. Defaults to
        string.ascii_lowercase+string.digits+string.ascii_uppercase.

    Returns:
        str: random string
    """
    return "".join(random.choice(alphabet) for _ in range(length))


def as_json(thing):
    """return thing as json.

    Args:
        thing (Any): thing.

    Returns:
        str: thing as json.
    """
    from abcli.file import JsonEncoder

    # https://docs.python.org/2/library/json.html
    return json.dumps(
        thing,
        sort_keys=True,
        cls=JsonEncoder,
        indent=0,
        ensure_ascii=False,
        separators=(",", ":"),
    ).replace("\n", "")


def shorten(thing, shorten_length=16):
    if isinstance(thing, list):
        return [shorten(item) for item in thing]
    return (
        thing
        if len(thing) < shorten_length
        else f"{thing[: shorten_length - 4]}..{thing[-2:]}"
    )


def timestamp():
    return pretty_date(
        as_filename=True,
        include_time=True,
        unique=True,
    )


def utc_timestamp(
    date=None,
    format="%Y-%m-%d-%H-%M-%S",
    timezone_="America/New_York",
):
    """generate utc timestamp.

    Args:
        date (Any, optional): date. Defaults to None.
        format (str, optional): format. Defaults to "%Y-%m-%d-%H-%M-%S".
        timezone_ (str, optional): time zone. Defaults to "America/New_York".

    Returns:
        str: utc timestamp.
    """

    if date is None:
        # https://www.geeksforgeeks.org/get-utc-timestamp-in-python/
        return (
            datetime.datetime.now(timezone.utc).replace(tzinfo=timezone.utc).timestamp()
        )

    try:
        # https://stackoverflow.com/a/79877/17619982
        import pytz

        local = pytz.timezone(timezone_)
        naive = datetime.datetime.strptime(date, format)
        local_dt = local.localize(naive, is_dst=None)
        return local_dt.astimezone(pytz.utc).timestamp()
    except:
        print(f"-{NAME}: utc_timestamp({date},{format}): failed.")
        return "unknown"



from typing import Tuple, Any, List
from copy import deepcopy
import json
import numpy as np
from abcli.file import NAME
from abcli.logger import logger, crash_report


def load(
    filename,
    civilized=False,
    default={},
) -> Tuple[bool, Any]:
    # https://wiki.python.org/moin/UsingPickle
    data = deepcopy(default)

    try:
        import dill

        with open(filename, "rb") as fp:
            data = dill.load(fp)

        return True, data
    except:
        if not civilized:
            crash_report(f"-{NAME}: load({filename}): failed.")

        return False, data


# https://stackoverflow.com/a/47792385/17619982
def load_geojson(
    filename,
    civilized=False,
) -> Tuple[bool, Any]:
    success = False
    data = {}

    try:
        import geojson

        with open(filename, "r") as fh:
            data = geojson.load(fh)

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_geojson({filename}): failed.")

    return success, data


def load_dataframe(
    filename,
    civilized=False,
    log=False,
) -> Tuple[bool, Any]:
    success = False
    df = None

    try:
        import pandas

        df = pandas.read_csv(filename)

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_dataframe({filename}): failed.")

    if success and log:
        logger.info(
            "loaded {} row(s) of {} from {}".format(
                len(df),
                ", ".join(df.columns),
                filename,
            )
        )

    return success, df


def load_geodataframe(
    filename,
    civilized=False,
) -> Tuple[bool, Any]:
    success = False
    gdf = None

    try:
        import geopandas

        gdf = geopandas.read_file(filename)

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_geodataframe({filename}): failed.")

    return success, gdf


def load_image(
    filename,
    civilized=False,
    log=False,
) -> Tuple[bool, np.ndarray]:
    import cv2
    from abcli import string

    success = True
    image = np.empty((0,))

    try:
        image = cv2.imread(filename)

        if len(image.shape) == 3:
            if image.shape[2] == 4:
                image = image[:, :, :3]

            image = np.flip(image, axis=2)

    except:
        if not civilized:
            crash_report(f"-{NAME}: load_image({filename}): failed.")
        success = False

    if success and log:
        logger.info(
            "loaded {} from {}".format(
                string.pretty_shape_of_matrix(image),
                filename,
            )
        )

    return success, image


def load_json(
    filename,
    civilized=False,
    default={},
) -> Tuple[bool, Any]:
    success = False
    data = default

    try:
        with open(filename, "r") as fh:
            data = json.load(fh)

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_json({filename}): failed.")

    return success, data


def load_text(
    filename,
    civilized=False,
    count=-1,
    log=False,
) -> Tuple[bool, List[str]]:
    success = True
    text = []

    try:
        if count == -1:
            with open(filename, "r") as fp:
                text = fp.read()
            text = text.split("\n")
        else:
            # https://stackoverflow.com/a/1767589/10917551
            with open(filename) as fp:
                text = [next(fp) for _ in range(count)]
    except:
        success = False
        if not civilized:
            crash_report(f"-{NAME}: load_text({filename}): failed.")

    if success and log:
        logger.info("loaded {} line(s) from {}.".format(len(text), filename))

    return success, text


def load_xml(
    filename,
    civilized=False,
    default={},
) -> Tuple[bool, Any]:
    success = False
    data = default

    try:
        import xml.etree.ElementTree as ET

        tree = ET.parse(filename)
        data = tree.getroot()

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_xml({filename}): failed.")

    return success, data


def load_yaml(
    filename,
    civilized=False,
    default={},
) -> Tuple[bool, Any]:
    success = False
    data = default

    try:
        import yaml

        with open(filename, "r") as fh:
            data = yaml.safe_load(fh)

        success = True
    except:
        if not civilized:
            crash_report(f"-{NAME}: load_yaml({filename}): failed.")

    return success, data
